import struct
from PIL import Image
import os
import glob
import sys

# fn - absolute filename
# an - absolute path with name (no extension)
# ad - absolute path to directory
# rn - relative filename

def binary(short):
	st = ""
	for ibit in range(15,-1,-1):
		if (1 << ibit) & short == 0:
			st = st + "0"
		else:
			st = st + "1"
	return st
	  
class Pic(object):
	"Holds a pil image and a filename, and can save the image."
	def __init__(self, image, ad, rn):
		self.image = image
		self.ad = ad
		self.rn = rn
		self.x = 0
		self.y = 0
		(self.w, self.h) = image.size
		
	def Save(self):
		print "Saving " + os.path.basename(self.Fn())
		if self.image.mode == "P":
			self.image.save(self.Fn(), transparency=0)
		else:
			self.image.save(self.Fn())
			
	def PutPixel(self, col):
		"Put a pixel in the image, autoincrementing x and y."
		if not self.FMore():
			self.Save()
			raise Exception("putpixel past end of picp!")
			
		self.image.putpixel((self.x, self.y), col)
		self.x = self.x + 1
		if self.x == self.w:
			self.x = 0
			self.y = self.y + 1
	
	def FMore(self):
		"Returns true if there are any more pixels to put."
		if self.y == self.h:
			return False
		return True
	
	def ImageMask(self):
		if self.image.mode == "P":
			imgMask = Image.new("1", (self.w, self.h))
			for x in range(self.w):
				for y in range(self.h):
					if self.image.getpixel((x, y)) == 0:
						imgMask.putpixel((x, y), 0)
					else:
						imgMask.putpixel((x, y), 1)
			return imgMask
		return self.image

	def Fn(self):
		return self.ad + "/" + self.rn + ".png"
		
class Pal(object):
	def __init__(self, fnPngPal):
		self.image = Image.open(fnPngPal)
	
	def Palette(self):
		"Generate a palette for pil (list of 768 integers) based on a colour grid."
		(w, h) = self.image.size
		wBox = w / 16
		hBox = h / 16
		palette = []
		# {x,y}p - "palette-based" xy value (0-15)
		for xp in range(16):
			for yp in range(16):
				palette.extend(self.image.getpixel((xp * wBox, yp * hBox)))
		return palette

class KnpExtractor(object):
	def __init__(self, an, pal):
		self.pal = pal
		self.an = an
		
	def FnImg(self):
		return self.an + ".img"
	
	def AdOut(self):
		return self.an
							
	def Extract(self):
		if not os.path.exists(self.AdOut()):
			os.mkdir(self.AdOut())
		for pic in self.Rgpic():
			pic.Save()
			
	def Rgpic(self):
		with open(self.FnImg(), "rb") as f:
			rfFile = f.read()
		(cimgHdr,) = struct.unpack_from("<I", rfFile);
		foImgHdr = 4
		for iimgHdr in range(cimgHdr):
			(foImg, szImg) = struct.unpack_from("<II", rfFile, foImgHdr)
			if foImg > 0:
				yield self.PicFromRfImg(iimgHdr, rfFile[foImg:foImg + szImg])
			foImgHdr = foImgHdr + 8
			
	def PicFromRfImg(self, iimg, rfImg):
		(width, height) = struct.unpack_from("<HH", rfImg, 10)
		(cmpf, hotX, hotY) = struct.unpack_from("<Bhh", rfImg, 15) # hotspots and action points can be negative(!)
		rn = str(iimg) + "_" + str(hotX) + "_" + str(hotY)
		if (width % 2) == 1:
			width = width + 1
		image = Image.new("P", (width, height), 0)
		image.putpalette(self.pal.Palette())
		pic = Pic(image, self.AdOut(), rn)
		rfImdat = rfImg[24:-1] # remove trailing zero
		
		if cmpf == 0:
			self.PutImageRaw(pic, rfImdat)
		elif cmpf == 1:
			self.PutImageRle(pic, rfImdat)
		else:
			raise Exception("Invalid compression format " + cmpf)
		
		return pic
		
	def PutImageRaw(self, pic, rfImdat):
		pic.image.putdata(rfImdat)
		
	def PutImageRle(self, pic, rfImdat):
		irfImdat = 0
		while pic.FMore():
			ctl = ord(rfImdat[irfImdat])
			irfImdat = irfImdat + 1
			if (ctl & 0x80) == 0: # rle
				cpixSame = ctl
				col = ord(rfImdat[irfImdat])
				for ipixSame in range(cpixSame):
					pic.PutPixel(col)
				irfImdat = irfImdat + 1
			else:
				cpixUnique = ctl & 0x7f
				for ipixUnique in range(cpixUnique):
					pic.PutPixel(ord(rfImdat[irfImdat]))
					irfImdat = irfImdat + 1

class KnpLoader(object):
	def __init__(self, adExe, rgarg):
		self.pal = Pal(adExe + "/knp_pal.png")
		self.rgarg = rgarg
	
	def RgExtractor(self):
		if (len(self.rgarg) == 0):
			for extractor in self.ScanAd(os.getcwd()):
				yield extractor
		else:
			for arg in self.rgarg:
				fnImg = self.AnFromFn(arg) + ".img"
				if os.path.exists(fnImg):
					yield self.Extractor(fnImg)
				elif os.path.isdir(arg):
					for extractor in self.ScanAd(arg):
						yield extractor
	
	
	def ScanAd(self, ad):
		for fnImg in glob.glob(ad + "/*.img"):
			yield self.Extractor(fnImg)

	def AnFromFn(self, fn):
		if fn.lower().endswith(".img"):
			return fn[:-4]
		return fn
	
	def Extractor(self, fn):
		return KnpExtractor(self.AnFromFn(fn), self.pal)
		
def main(adExe, rgarg):
	print "KNPExtract v0.1b"
	print "By Jeremy Penner"
	print "This code is public domain"
	print
	print "Usage:"
	print "KNPExtract [directory to scan for .img files | .img file] ..."

	ldr = KnpLoader(adExe, rgarg)
	for extractor in ldr.RgExtractor():
		print "Extracting from " + os.path.basename(extractor.FnImg()) + "..."
		try:
			extractor.Extract()
		except e:
			print "Error: " + e

def AdExe(fnExe):
	return os.path.dirname(os.path.abspath(sys.argv[0]))
	
if __name__ == "__main__":
	main(AdExe(sys.argv[0]), sys.argv[1:])